@@ -23,6 +23,7 @@
 #include <gtk/gtk.h>
 
 #include <libkoto/ical-util.h>
+#include <libkoto/koto-actions.h>
 #include <libkoto/koto-category-group.h>
 #include <libkoto/koto-all-group.h>
 #include <libkoto/koto-no-category-group.h>
@@ -37,6 +38,8 @@
 #include <libkoto/koto-group-combo.h>
 #include <libkoto/koto-platform.h>
 #include <libkoto/koto-utils.h>
+#include <libkoto/koto-undo-action.h>
+#include <libkoto/koto-undo-manager.h>
 
 #include "window-util.h"
 
@@ -46,6 +49,7 @@ static GtkTreeModel *task_store, *filter, *group_filter_store, *group_selector_s
 /* Hash of KotoTask to KotoTaskEditorDialogs, to avoid opening the same editor
    twice */
 static GHashTable *edit_dialog_hash;
+static KotoUndoManager *undo_manager;
 
 void
 koto_platform_open_url (const char *url)
@@ -175,11 +179,11 @@ create_component (const char *text, KotoGroup *default_group)
 static void
 on_new_clicked (GtkButton *button)
 {
-  GError *error = NULL;
   const char *text;
   char *uid = NULL;
   KotoGroup *group;
   icalcomponent *comp;
+  KotoUndoContext *ctxt;
 
   text = gtk_entry_get_text (GTK_ENTRY (new_entry));
   
@@ -192,12 +196,16 @@ on_new_clicked (GtkButton *button)
   comp = create_component (text, group);
   g_object_unref (group);
 
-  koto_hint_entry_clear (KOTO_HINT_ENTRY (new_entry));
+  ctxt = koto_undo_manager_context_begin_formatted (undo_manager,
+                                                    _("Add task %s"),
+                                                    icalcomponent_get_summary (comp),
+                                                    NULL);
 
-  if (!e_cal_create_object (cal, comp, &uid, &error)) {
-    g_warning (G_STRLOC ": cannot create task: %s", error->message);
-    g_error_free (error);
-  }
+  koto_action_create_task (cal, comp, &uid, ctxt);
+  
+  koto_undo_manager_context_end (undo_manager, ctxt);
+
+  koto_hint_entry_clear (KOTO_HINT_ENTRY (new_entry));
 
   /*
    * Select the new task in an idle function so that the store can process the
@@ -209,7 +217,6 @@ on_new_clicked (GtkButton *button)
   }
 }
 
-
 /* 
  * Used to enable/disable the new task button depending on the contents of the
  * entry.
@@ -246,6 +253,7 @@ edit_task (KotoTask *task)
                   "cal", cal,
                   "groups", group_selector_store,
                   "task", task,
+                  "undo-manager", undo_manager,
                   NULL);
     
     gtk_widget_show (dialog);
@@ -270,13 +278,24 @@ on_row_activated  (GtkTreeView *tree_view, GtkTreePath *path, GtkTreeViewColumn
     g_warning (G_STRLOC ": cannot get iterator for path");
     return;
   }
-  gtk_tree_model_get (model, &iter, COLUMN_ICAL, &task, -1);
+  gtk_tree_model_get (model, &iter, COLUMN_TASK, &task, -1);
 
   edit_task (task);
 
   koto_task_unref (task);
 }
 
+/*
+ * Callback when the selection changes, to disable some actions */
+static void
+on_selection_changed (GtkTreeSelection *selection, gpointer user_data)
+{
+  GtkActionGroup *actions = user_data;
+
+  gtk_action_group_set_sensitive
+    (actions, gtk_tree_selection_count_selected_rows (selection) != 0);
+}
+
 /*
  * Callback from the New Task action.
  */
@@ -296,7 +315,7 @@ on_edit_task_action (GtkAction *action, gpointer user_data)
 
   task = koto_task_view_get_selected_task (KOTO_TASK_VIEW (treeview));
   if (!task) {
-    g_warning ("TODO: No task selected, EditTask should be disabled");
+    g_warning ("No task selected, EditTask should be disabled");
     return;
   }
   
@@ -309,15 +328,18 @@ static void
 on_complete_task_action (GtkAction *action, gpointer user_data)
 {
   GtkTreeIter real_iter, iter;
+  KotoUndoContext *undo;
 
   if (!koto_task_view_get_selected_iter (KOTO_TASK_VIEW (treeview), &iter)) {
-    g_warning ("TODO: No task selected, CompleteTask should be disabled");
+    g_warning ("No task selected, CompleteTask should be disabled");
     return;
   }
 
   gtk_tree_model_filter_convert_iter_to_child_iter (GTK_TREE_MODEL_FILTER (filter), &real_iter, &iter);
   
-  koto_task_store_set_done (KOTO_TASK_STORE (task_store), &real_iter, TRUE);
+  undo = koto_undo_manager_context_begin (undo_manager, _("Complete Task"));
+  koto_task_store_set_done (KOTO_TASK_STORE (task_store), &real_iter, TRUE, undo);
+  koto_undo_manager_context_end (undo_manager, undo);
 }
 
 /*
@@ -326,38 +348,22 @@ on_complete_task_action (GtkAction *action, gpointer user_data)
 static void
 on_delete_task_action (GtkAction *action, gpointer user_data)
 {
-  GError *error = NULL;
-  GtkWidget *dialog;
   KotoTask *task;
+  KotoUndoContext *ctxt;
 
   task = koto_task_view_get_selected_task (KOTO_TASK_VIEW (treeview));
   if (!task) {
-    g_warning ("TODO: No task selected, DeleteTask should be disabled");
+    g_warning ("No task selected, DeleteTask should be disabled");
     return;
   }
   
-  dialog = gtk_message_dialog_new (GTK_WINDOW (window),
-                                   GTK_DIALOG_DESTROY_WITH_PARENT,
-                                   GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
-                                   _("Are you sure you want to delete \"%s\"?"),
-                                   icalcomponent_get_summary (task->comp));
-  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
-                                            _("If you delete an item, it is permanently lost."));
-  gtk_dialog_add_buttons (GTK_DIALOG (dialog),
-                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
-                          GTK_STOCK_DELETE, GTK_RESPONSE_ACCEPT,
-                          NULL);
-  gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
+  ctxt = koto_undo_manager_context_begin_formatted (undo_manager,
+                                                    _("Delete Task %s"),
+                                                    icalcomponent_get_summary (task->comp),
+                                                    NULL);
+  koto_action_delete_task (cal, task, ctxt);
+  koto_undo_manager_context_end (undo_manager, ctxt);
   
-  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
-    if (!e_cal_remove_object (cal, icalcomponent_get_uid (task->comp), &error)) {
-      g_warning ("Cannot remove object: %s", error->message);
-      g_error_free (error);
-    }
-  }
-  
-  gtk_widget_destroy (dialog);
-
   koto_task_unref (task);
 }
 
@@ -368,21 +374,17 @@ on_delete_task_action (GtkAction *action, gpointer user_data)
 static gboolean
 purge_foreach (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
 {
-  GError *error = NULL;
+  KotoUndoContext *ctxt = data;
   KotoTask *task;
   gboolean done;
 
   gtk_tree_model_get (model, iter,
                       COLUMN_DONE, &done,
-                      COLUMN_ICAL, &task,
+                      COLUMN_TASK, &task,
                       -1);
   
   if (done) {
-    if (!e_cal_remove_object (cal, icalcomponent_get_uid (task->comp), &error)) {
-      g_warning ("Cannot remove object: %s", error->message);
-      g_error_free (error);
-      error = NULL;
-    }
+    koto_action_delete_task (cal, task, ctxt);
   }
 
   koto_task_unref (task);
@@ -396,25 +398,13 @@ purge_foreach (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpoint
 static void
 on_purge_action (GtkAction *action, gpointer user_data)
 {
-  GtkWidget *dialog;
+  KotoUndoContext *ctxt;
 
-  dialog = gtk_message_dialog_new (GTK_WINDOW (window),
-                                   GTK_DIALOG_DESTROY_WITH_PARENT,
-                                   GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
-                                   _("Are you sure you want to delete all completed tasks?"));
-  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
-                                            _("Deleting completed tasks means they are permanently lost."));
-  gtk_dialog_add_buttons (GTK_DIALOG (dialog),
-                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
-                          GTK_STOCK_DELETE, GTK_RESPONSE_ACCEPT,
-                          NULL);
-  gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
+  ctxt = koto_undo_manager_context_begin (undo_manager, _("Remove Completed"));
   
-  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
-    gtk_tree_model_foreach (task_store, purge_foreach, NULL);
-  }
+  gtk_tree_model_foreach (task_store, purge_foreach, ctxt);
   
-  gtk_widget_destroy (dialog);
+  koto_undo_manager_context_end (undo_manager, ctxt);
 }
 
 /*
@@ -468,15 +458,21 @@ static const GtkActionEntry actions[] =
   /* Action name, stock ID, label, accelerator, tooltip, callback */
   { "TasksMenu", NULL, N_("_Task") },
   { "NewTask", GTK_STOCK_NEW, NULL, NULL, NULL, G_CALLBACK (on_new_task_action) },
+  { "PurgeTasks", NULL, N_("_Remove Completed"), NULL, NULL, G_CALLBACK (on_purge_action) },
+  { "Quit", GTK_STOCK_QUIT, NULL, NULL, NULL, G_CALLBACK (gtk_main_quit) },
+
+  { "EditMenu", NULL, N_("_Edit") },
+
+  { "HelpMenu", NULL, N_("_Help") },
+  { "About", GTK_STOCK_ABOUT, NULL, NULL, NULL, G_CALLBACK (on_about_action) },
+};
+
+static const GtkActionEntry task_actions[] = 
+{
   { "EditTask", GTK_STOCK_EDIT, NULL, "<control>e", NULL, G_CALLBACK (on_edit_task_action) },
   /* TODO: turn this action into a toggle action */
   { "CompleteTask", NULL, N_("_Mark Complete"), "<control>d", NULL, G_CALLBACK (on_complete_task_action) },
   { "DeleteTask", GTK_STOCK_DELETE, NULL, "<control>Delete", NULL, G_CALLBACK (on_delete_task_action) },
-  { "PurgeTasks", NULL, N_("_Remove Completed"), NULL, NULL, G_CALLBACK (on_purge_action) },
-  { "Quit", GTK_STOCK_QUIT, NULL, NULL, NULL, G_CALLBACK (gtk_main_quit) },
-  
-  { "HelpMenu", NULL, N_("_Help") },
-  { "About", GTK_STOCK_ABOUT, NULL, NULL, NULL, G_CALLBACK (on_about_action) },
 };
 
 int
@@ -485,9 +481,10 @@ main (int argc, char **argv)
   GError *error = NULL;
   ECalView *cal_view;
   GtkWidget *top_box, *box, *menu, *scrolled, *hbox, *label, *new_button;
-  GtkActionGroup *action_group;
+  GtkTreeSelection *selection;
+  GtkActionGroup *action_group, *task_action_group;
   GtkUIManager *ui_manager;
-
+  
 #ifdef ENABLE_NLS
   /* Initialise i18n*/
   bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
@@ -500,6 +497,8 @@ main (int argc, char **argv)
   
   edit_dialog_hash = g_hash_table_new_full (NULL, NULL, (GDestroyNotify)koto_task_unref, NULL);
 
+  undo_manager = koto_undo_manager_new ();
+
   cal = e_cal_new_system_tasks ();
   if (!cal)
     g_error ("Cannot get system tasks");
@@ -540,9 +539,20 @@ main (int argc, char **argv)
   action_group = gtk_action_group_new ("Actions");
   gtk_action_group_set_translation_domain (action_group, GETTEXT_PACKAGE);
   gtk_action_group_add_actions (action_group, actions, G_N_ELEMENTS (actions), NULL);
+  gtk_action_group_add_action_with_accel (action_group,
+                                          koto_undo_action_new ("Undo", undo_manager),
+                                          "<Control>z");
+  gtk_action_group_add_action_with_accel (action_group,
+                                          koto_undo_action_new_redo ("Redo", undo_manager),
+                                          "<Control><Shift>z");
+
+  task_action_group = gtk_action_group_new ("Task Actions");
+  gtk_action_group_set_translation_domain (task_action_group, GETTEXT_PACKAGE);
+  gtk_action_group_add_actions (task_action_group, task_actions, G_N_ELEMENTS (task_actions), NULL);
 
   ui_manager = gtk_ui_manager_new ();
   gtk_ui_manager_insert_action_group (ui_manager, action_group, 0);
+  gtk_ui_manager_insert_action_group (ui_manager, task_action_group, 0);
 
   gtk_ui_manager_add_ui_from_file (ui_manager, PKGDATADIR "/tasks-ui.xml", &error);
   if (error) {
@@ -577,11 +587,16 @@ main (int argc, char **argv)
   filter = koto_group_model_filter_new (KOTO_TASK_STORE (task_store));
   g_object_unref (task_store);
   treeview = koto_task_view_new (KOTO_TASK_STORE (task_store), KOTO_GROUP_MODEL_FILTER (filter));
+  g_object_set (treeview, "undo-manager", undo_manager, NULL);
   g_object_unref (filter);
 
   g_signal_connect (treeview, "row-activated", G_CALLBACK (on_row_activated), NULL);
   gtk_container_add (GTK_CONTAINER (scrolled), treeview);
 
+  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
+  g_signal_connect (selection, "changed", G_CALLBACK (on_selection_changed), task_action_group);
+  on_selection_changed (selection, task_action_group);
+  
   hbox = gtk_hbox_new (FALSE, 4);
   new_entry = koto_hint_entry_new (_("New task..."));
   gtk_entry_set_activates_default (GTK_ENTRY (new_entry), TRUE);
