@@ -1,7 +1,7 @@
 /*
- * Original Copyright (C) 2007 OpenedHand Ltd
+ * Copyright (C) 2007 OpenedHand Ltd.
  * Derived from gtk/main.c.
- * Derivations (C) Rob Bradford <rob@o-hand.com> 2007
+ * Authored by: Rob Bradford <rob@o-hand.com>
  *
  * This program is free software; you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free Software
@@ -246,7 +246,7 @@ on_row_activated  (GtkTreeView *tree_view, GtkTreePath *path, GtkTreeViewColumn
     g_warning (G_STRLOC ": cannot get iterator for path");
     return;
   }
-  gtk_tree_model_get (model, &iter, COLUMN_ICAL, &task, -1);
+  gtk_tree_model_get (model, &iter, COLUMN_TASK, &task, -1);
 
   edit_task (task);
 
@@ -350,7 +350,7 @@ purge_foreach (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpoint
 
   gtk_tree_model_get (model, iter,
                       COLUMN_DONE, &done,
-                      COLUMN_ICAL, &task,
+                      COLUMN_TASK, &task,
                       -1);
   
   if (done) {
